// -*-c++-*-

/***************************************************************************
                          rccangle.h  -  Interface for a general purpose
			                 angle class.
                             -------------------
    begin                : 17-JULY-2001
    copyright            : (C) 2001 by Tom Howard.
    email                : thomashoward@email.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef _RCC_ANGLE_H_
#define _RCC_ANGLE_H_

#pragma interface

#include <math.h>
#include "rccobject.h"
#include "rccmemory.h"
#include "rccstring.h"

// comment out the following line if  is not displayed as degree symbol on your system
#define DEGREE_SYMBOL ""

#ifndef DEGREE_SYMBOL
#define DEGREE_SYMBOL "^"
#endif

class RCC_Angle : public RCC_Object
{
public:
  enum mode_t { DEG, RAD, BOTH };
    
  RCC_Angle ();				// default constructor. Set Invalid
  RCC_Angle ( const double&, const mode_t& ); 	// constuct an RCC_Angle in specifed mode    
    
  RCC_Pointer < RCC_Object > copy () const { return new RCC_Angle ( *this ); }

  RCC_String toString ( const RCC_Angle::mode_t& m = RCC_Angle::DEG ) const;

  double rad() const { if ( valid ) { toRad(); return rad_value; } return HUGE_VAL; }		// returns RCC_Angle in rads
  double deg() const { if ( valid ) { toDeg(); return deg_value; } return HUGE_VAL; }		// returns RCC_Angle in degs

  bool isValid() const { return valid; }
  RCC_Angle& clear() { valid = false; return *this; }
        
  const RCC_Angle& norm();			// normalises the RCC_Angle
  const RCC_Angle& trunc();			// truncates the RCC_Angle

  double cos() const { return ::cos ( rad () ); }
  double sin() const { return ::sin ( rad () ); }
  double tan() const { return ::tan ( rad () ); }

  RCC_Angle& operator+= (const RCC_Angle& ang);
  RCC_Angle& operator-= (const RCC_Angle& ang);
   
  RCC_Angle& operator*= (const double& s);
  RCC_Angle& operator/= (const double& s);

  RCC_Angle operator-() const;
  RCC_Angle operator+() const { return *this; }

  friend RCC_Angle operator+(RCC_Angle a, const RCC_Angle& d) { return (a += d); }
  friend RCC_Angle operator-(RCC_Angle a, const RCC_Angle& d) { return (a -= d); }

  friend RCC_Angle operator*(const double& d, RCC_Angle a) { return (a *= d); }
  friend RCC_Angle operator*(RCC_Angle a, const double& d) { return (a *= d); }
    
  friend RCC_Angle operator/(RCC_Angle a, double d) { return (a /= d); }

  friend bool equals(const RCC_Angle&, const RCC_Angle&);
  friend bool operator== ( const RCC_Angle &a, const RCC_Angle &b ) { return ( equals ( a, b ) ); }
    
  friend std::ostream& operator<< (std::ostream& o, const RCC_Angle& ang) { return ang.print ( o ); }
  std::ostream& print ( std::ostream& o ) const;
  std::ostream& print ( std::ostream& o, RCC_Angle::mode_t m ) const; 
  std::ostream& printPretty ( std::ostream& o, RCC_Angle::mode_t m ) const; 

  static RCC_Angle asin(const double &i);
  static RCC_Angle acos(const double &i);
  static RCC_Angle atan(const double &i);
  static RCC_Angle atan2(const double &y, const double& x);

private:
  bool valid;
  mutable double deg_value;
  mutable double rad_value;
  mutable mode_t mode;
  
  static double rad2deg ( const double& rad ) { return ( rad * 180 * M_1_PI ); }	// converts from radians to degrees
  static double deg2rad ( const double& deg ) { return ( deg * M_PI / 180.0 ); }	// converts from degrees to radians

  static double normDeg ( double deg )
  {
    if ( deg > 180.0 || deg < -180.0 )
      {
	if ( fabs ( deg ) > 720.0 )
	  {
	    int tmp = ( int ) deg / 360;
	    deg -= 360 * tmp;
	  }

	while ( deg > 180.0 )
	  deg -= 360.0;
	while ( deg < -180.0 )
	  deg += 360.0;
      }
    return deg;
  }

  static double normRad ( double rad )
  {
    if ( rad > M_PI || rad < -M_PI )
      {
	if ( fabs ( rad ) > M_PI * 4.0 )
	  {
	    int tmp = ( int ) floor ( rad / ( M_PI * 2.0 ) );
	    rad -= M_PI * 2.0 * tmp;
	  }

	while ( rad > M_PI )
	  rad -= M_PI * 2.0;
	while ( rad < -M_PI )
	  rad += M_PI * 2.0;
      }
    return rad;
  }

  static double truncDeg ( const double& deg )
  {
    if ( deg > 180.0 )
      return 180.0;
    else if ( deg < -180.0 )
      return -180.0;
    else
      return deg;
  }

  static double truncRad ( const double& rad )
  {
    if ( rad > M_PI )
      return M_PI;
    else if ( rad < -M_PI )
      return -M_PI;
    else
      return rad;
  }

    
        
  const RCC_Angle& toRad() const // converts to radians and returns itself
  {
    if ( mode == RCC_Angle::DEG )
      {
	rad_value = RCC_Angle::deg2rad ( deg_value );
	mode = RCC_Angle::BOTH;
      }
    return *this;
  }
		
  const RCC_Angle& toDeg() const		// converts to degrees and returns itself
  {
    if ( mode == RCC_Angle::RAD )
      {
	deg_value = rad2deg ( rad_value );
	mode = RCC_Angle::DEG;
      }
    return *this;
  }

  RCC_Angle ( const double& );	// not allowed  
};


inline double cosineRule(const RCC_Angle& a, const double& B, const double& C)
{
  return sqrt( B*B + C*C - 2.0 * B * C * a.cos() );
}

inline RCC_Angle cosineRule(const double& A, const double& B, const double& C)
{
  return RCC_Angle::acos( ( B*B + C*C - A*A ) / (2.0 * B * C) );
}


#endif
