// -*-c++-*-

/***************************************************************************
                          rccmemory.h  -  Interface for a reference counting
			                  class and smart pointers
                             -------------------
    begin                : 17-JULY-2001
    copyright            : (C) 2001 by Tom Howard.
    email                : thomashoward@email.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef _RCC_MEMORY_H_
#define _RCC_MEMORY_H_

#pragma interface

#include <map>
#include <iostream>
#include <vector>
#include <typeinfo>
#include "rccobject.h"
#include "rccexception.h"
#include "rcclockable.h"


class RCC_Memory 
  : protected RCC_Lockable
{
public:
  typedef std::map < const RCC_Object*, unsigned long > container;
  static RCC_Memory DEFAULT_MEM;

  RCC_Memory () {}
  ~RCC_Memory () { delAll (); }

  void delAll ()
  {
    lock ();
    for ( RCC_Memory::container::iterator i = mem.begin ();
	  i != mem.end (); ++i )
      {
	unlock ();
	delete i->first;
	lock ();
      }
    mem.clear ();
    unlock ();
  }

  void add ( const RCC_Object* ptr )
  {
    if ( ptr )
      {
	lock ();
  	if ( mem.find ( ptr ) == mem.end () )
  	  mem.insert ( std::pair < const RCC_Object*, unsigned long > ( ptr, 1 ) );
  	else
	  {
	    mem [ ptr ]++;
	    if ( mem [ ptr ] == 0 )
	      {
		unlock ();
		throw RCC_MemoryException ( "Reference count overflow" );
	      }
	  }
	unlock ();
      }
  }

  void remove ( RCC_Object* ptr )
  {
    if ( ptr )
      {
	lock ();
	RCC_Memory::container::iterator iter = mem.find ( ptr );
	if ( iter != mem.end () )
	  {
	    (iter->second)--;
	    if ( iter->second <= 0 )
	      {
		mem.erase ( iter );
		unlock ();
		delete ptr;
		ptr = NULL;
		return;
	      }
	  }    
	unlock ();
      }
  }

  // using mem [ ptr ] will add ptr to mem if it's not found
  // so we use find instead to avoid this.
  unsigned long count ( const RCC_Object* ptr ) const 
  { 
    lock ();
    RCC_Memory::container::const_iterator iter = mem.find ( ptr );
    unsigned long tmp = 0;
    if ( iter != mem.end () )
	tmp = iter->second;
    unlock ();
    return tmp;
  }

  friend std::ostream& operator << ( std::ostream& o, const RCC_Memory& m )
  {
    m.lock ();
    o << typeid ( m ).name () << " ( ";
    for ( RCC_Memory::container::const_iterator i = m.mem.begin ();
	  i != m.mem.end (); ++i )
      {
	o << i->first << "=" << i->second << " ";
      }
    m.unlock ();
    return o << ")";
  }

  unsigned int size () const 
  { 
    lock ();
    unsigned int tmp = mem.size (); 
    unlock (); 
    return tmp; 
  }

private:
  RCC_Memory::container mem;
};



template < class T > class RCC_Pointer
{
public:

  RCC_Pointer< T > ( T* p = NULL, RCC_Memory& m = RCC_Memory::DEFAULT_MEM ) : ptr_ ( p ) , mem ( m )
  {
    mem.add ( ptr_ );
  }

  RCC_Pointer< T > ( const RCC_Pointer < T >& p ) : ptr_ ( p.ptr_ ), mem ( p.mem )
  {
    mem.add ( ptr_ );
  }

  ~RCC_Pointer< T > () 
  {
    mem.remove ( ptr_ );
  }

  RCC_Pointer< T >& operator= ( const RCC_Pointer < T >& p )
  {
    mem.remove ( ptr_ );
    ptr_ = p.ptr_;
    mem = p.mem;
    mem.add ( ptr_ );
    return *this;
  }	


  T*
  operator-> () 
  { 
    if ( ptr_ ) 
      return ptr_; 
    throw RCC_MemoryException ( "Attempt to dereference a null pointer" );
  }

  const T* 
  operator-> () const 
  { 
    if ( ptr_ ) 
      return ptr_; 
    throw RCC_MemoryException ( "Attempt to dereference a null pointer" );
  }

  T& 
  operator* () 
  { 
    if ( ptr_ ) 
      return *ptr_; 
    throw RCC_MemoryException ( "Attempt to dereference a null pointer" );
  }


  const T& 
  operator* () const 
  { 
    if ( ptr_ ) 
      return *ptr_; 
    throw RCC_MemoryException ( "Attempt to dereference a null pointer" );
  }
  
  operator const T* () const { return ptr_; }
  operator T* () { return ptr_; }

  template < class T2 > 
  operator RCC_Pointer < T2 > () 
  { return RCC_Pointer < T2 > ( ptr_, mem ); }
  
  template < class T2 > 
  RCC_Pointer < T >& 
  dynamicCast ( RCC_Pointer < T2 > t ) 
  {
    mem.remove ( ptr_ );
    ptr_ = dynamic_cast < T* > ( ( T2* ) t );
    mem = t.memory ();
    mem.add ( ptr_ );
    return *this;
  }

  template < class T2 > 
  RCC_Pointer < T >& 
  staticCast ( RCC_Pointer < T2 > t ) 
  {
    mem.remove ( ptr_ );
    ptr_ = static_cast < T* > ( ( T2* ) t );
    mem = t.memory ();
    mem.add ( ptr_ );
    return *this;
  }
  
  template < class T2 > 
  RCC_Pointer < T2 > 
  dynamicCast () const 
  {
    return dynamic_cast < T2* > ( ptr_ );
  }

  template < class T2 > 
  RCC_Pointer < T2 > 
  staticCast () const 
  {
    return static_cast < T2* > ( ptr_ );
  }


  RCC_Memory& 
  memory () const 
  { return mem; }

private:
  T* ptr_;
  mutable RCC_Memory& mem;
};

#endif
