// -*-c++-*-

/***************************************************************************
                          rccvisualobject.h  -  Interface for RoboCup
			                        objects found is visual
			                        (or "see") messages.
                             -------------------
    begin                : 17-JULY-2001
    copyright            : (C) 2001 by Tom Howard.
    email                : thomashoward@email.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef _RCC_VISUAL_OBJECT_H_
#define _RCC_VISUAL_OBJECT_H_

#pragma interface

#include "rccstring.h"
#include <strstream>
#include "rccvector2d.h"
#include "rccangle.h"
#include <vector>
#include "rccmemory.h"
#include "rccobject.h"
#include "rcccontainer.h"

//=======================================================================
//	RCC_Object
//=======================================================================

class RCC_VisualObject : public RCC_Object 
{
public:
  typedef RCC_Container < RCC_VisualObject > container;
  typedef RCC_Container < const RCC_Object > const_container;

  enum location_t { 
    INVALID_LOC,
    CLOSE_LOC,
    TOP_LOC, 
    BOTTOM_LOC, 
    LEFT_LOC, 
    RIGHT_LOC, 
    CENTER_LOC, 
    PENALTY_LOC, 
    GOAL_LOC,
    CLOSE_GOAL_LOC,
    MAX_LOC 
  };

  static const RCC_String LOCATION_NAMES[];

  RCC_VisualObject ( const RCC_Vector2d& pos );
  RCC_VisualObject ( const RCC_Angle& dir );
  virtual ~RCC_VisualObject();
   
  RCC_Vector2d relPos() const { return rel_pos; }
  RCC_Vector2d relPos ( const RCC_Vector2d &pos ) { return ( rel_pos = pos ); }
      
  friend std::ostream& operator<< ( std::ostream& o, const RCC_VisualObject& obj ) { return obj.print ( o ); }

  virtual std::ostream& print ( std::ostream& o ) const
  { return print ( o, RCC_Vector2d::POLAR, RCC_Angle::DEG ); }
  virtual std::ostream& print ( std::ostream& o, RCC_Vector2d::mode_t, RCC_Angle::mode_t ) const = 0;

private:
  RCC_Vector2d rel_pos;		// objects relative position to player
};

//=======================================================================
//	RCC_StatObject
//=======================================================================

class RCC_StatObject : public RCC_VisualObject
{
public:
  RCC_StatObject ( const RCC_Vector2d& pos );
  RCC_StatObject ( const RCC_Angle& dir );
  virtual ~RCC_StatObject();
};

//=======================================================================
//	RCC_Line
//=======================================================================


class RCC_Line : public RCC_StatObject
{
public:
  typedef RCC_Container < RCC_Line > container;
	
  RCC_Line ( const RCC_VisualObject::location_t& name, 
	     const RCC_Angle &dir );
  
  RCC_Line ( const RCC_VisualObject::location_t& name, 
	     const RCC_Vector2d &pos );  

  ~RCC_Line();

  RCC_Pointer < RCC_Object >
  copy () const 
  { return new RCC_Line ( *this ); }

  RCC_VisualObject::location_t 
  name () const 
  { return l_name; }
  
  RCC_VisualObject::location_t
  name ( const RCC_VisualObject::location_t& ln )
  { return ( l_name = ln ); }
    
  virtual std::ostream& print ( std::ostream& o, RCC_Vector2d::mode_t = RCC_Vector2d::POLAR, RCC_Angle::mode_t = RCC_Angle::DEG ) const;

private:
  RCC_VisualObject::location_t l_name;
};

//=======================================================================
//	RCC_MobObject
//=======================================================================

class RCC_MobObject : public RCC_VisualObject
{
public:
  static RCC_Vector2d vel(const RCC_Vector2d &pos, const RCC_Vector2d &chg);
  
  RCC_MobObject ( const RCC_Angle& dir );
  RCC_MobObject ( const RCC_Vector2d& pos );
  RCC_MobObject ( const RCC_Vector2d& pos, const RCC_Vector2d& chg ); 
  virtual ~RCC_MobObject();

  RCC_Vector2d relVel() const { return rel_vel; }
  RCC_Vector2d relVel( const RCC_Vector2d v ) { return ( rel_vel = v ); }
    
  RCC_Vector2d distDirChg() const;
    
private:
  RCC_Vector2d rel_vel;
};

//=======================================================================
//	RCC_Flag
//=======================================================================

// Note:  	I've decided that flag should inherit from RCC_MobObject, rather than RCC_StatObject
//		as it can have DistChg and DirChg from the players own movement
//		Since a flag can have a relative velocity, it should be a RCC_MobObject...
//		...though I might change my mind in the future :)
 
class RCC_Flag : public RCC_MobObject
{
public: 
  typedef RCC_Container < RCC_Flag > container;

  RCC_Flag( const int &name, const RCC_Angle& dir ); 
  RCC_Flag( const int &name, const RCC_Vector2d &pos ); 
  RCC_Flag( const int &name, const RCC_Vector2d &pos, const RCC_Vector2d &chg ); 
  ~RCC_Flag();

  RCC_Pointer < RCC_Object > copy () const { return new RCC_Flag ( *this ); }

  int name () const { return f_name; }
  int name ( const int& fn ) { return ( f_name = fn ); }
    
  virtual std::ostream& print ( std::ostream& o, RCC_Vector2d::mode_t = RCC_Vector2d::POLAR, RCC_Angle::mode_t = RCC_Angle::DEG ) const;

  bool isGoal () const 
  { 
    return ( name () == RCC_Flag::getNum ( RCC_VisualObject::CLOSE_GOAL_LOC ) ||
	     name () == RCC_Flag::getNum ( RCC_VisualObject::GOAL_LOC, RCC_VisualObject::LEFT_LOC ) ||
	     name () == RCC_Flag::getNum ( RCC_VisualObject::GOAL_LOC, RCC_VisualObject::RIGHT_LOC ) ); 
  }

  RCC_String getString () const;
  static int getNum ();
  static int getNum ( const RCC_VisualObject::location_t& l);
  static int getNum ( const RCC_VisualObject::location_t& l1, const RCC_VisualObject::location_t& l2 );
  static int getNum ( const RCC_VisualObject::location_t& l1, const int& l2 );
  static int getNum ( const RCC_VisualObject::location_t& l1, const RCC_VisualObject::location_t& l2, const int& l3 );
  static int getNum ( const RCC_VisualObject::location_t& l1, const RCC_VisualObject::location_t& l2, const location_t& l3 );

private:
  int f_name;    
};


//=======================================================================
//	RCC_Ball
//=======================================================================

class RCC_Ball : public RCC_MobObject
{
public:
  typedef RCC_Container < RCC_Ball > container;

  RCC_Ball ( const bool& c, const RCC_Angle &dir );
  RCC_Ball ( const bool& c, const RCC_Vector2d &pos );
  RCC_Ball ( const bool& c, const RCC_Vector2d &pos, const RCC_Vector2d &chg );
  ~RCC_Ball();

  RCC_Pointer < RCC_Object > copy () const { return new RCC_Ball ( *this ); }

  virtual std::ostream& print ( std::ostream& o, RCC_Vector2d::mode_t = RCC_Vector2d::POLAR, RCC_Angle::mode_t = RCC_Angle::DEG ) const;

  bool close () const { return close_; }
  bool close ( const bool& c ) { return close_ = c; }

private:
  bool close_;
};

//=======================================================================
//	RCC_PlayerName
//=======================================================================

class RCC_PlayerName : public RCC_Object
{
public:
  static const int UNKNOWN_NUM;
  static const RCC_String UNKNOWN_TEAM;

  RCC_PlayerName ( const RCC_String& t, const int& u, const bool& g ) 
    : team_name ( t ), 
    unum ( u ), 
    goal_keeper ( g ),
    close_ ( false )
  {}

  RCC_PlayerName ( const RCC_String& t, const int& u ) 
    : team_name ( t ), 
    unum ( u ), 
    goal_keeper ( false ),
    close_ ( false )
  {}

  RCC_PlayerName ( const RCC_String& t ) 
    : team_name ( t ), 
    unum ( RCC_PlayerName::UNKNOWN_NUM ), 
    goal_keeper ( false ),
    close_ ( false )
  {}

  RCC_PlayerName ( const bool& c ) 
    : team_name ( RCC_PlayerName::UNKNOWN_TEAM ), 
    unum ( RCC_PlayerName::UNKNOWN_NUM ), 
    goal_keeper ( false ),
    close_ ( c )
  {}

  RCC_Pointer < RCC_Object > copy () const { return new RCC_PlayerName ( *this ); }

  bool isKnown () const { return team_name != RCC_PlayerName::UNKNOWN_TEAM; }

  RCC_String team() const { return team_name; }
  RCC_String team( const RCC_String& tname ) { return ( team_name = tname ); }

  bool goalie () const { return goal_keeper; }
  bool goalie ( const bool &gk ) { return ( goal_keeper = gk ); }
    
  int uNum() const { return unum; }
  int uNum( const int& num ) { return ( unum = num ); }
    
  friend std::ostream& operator<< ( std::ostream& o, const RCC_PlayerName& pn ) { return pn.print ( o ); }
  std::ostream& print ( std::ostream& o ) const;

  bool close () const { return close_; }
  bool close ( const bool &c ) { return ( close_ = c ); }
 
private:
  RCC_String team_name;
  int unum;
  bool goal_keeper;
  bool close_;
};

//=======================================================================
//	RCC_Player
//=======================================================================

class RCC_Player : public RCC_MobObject
{
public:
  typedef RCC_Container < RCC_Player > container;
    
  RCC_Player ( const RCC_PlayerName& pn, const RCC_Angle& dir ); 
  RCC_Player ( const RCC_PlayerName& pn, const RCC_Vector2d &pos ); 
  RCC_Player ( const RCC_PlayerName& pn, const RCC_Vector2d &pos, const RCC_Vector2d &chg ); 
  RCC_Player ( const RCC_PlayerName& pn, const RCC_Vector2d &pos, const RCC_Vector2d &chg, const RCC_Angle &body, const RCC_Angle &head ); 
  ~RCC_Player();

  RCC_Pointer < RCC_Object > copy () const { return new RCC_Player ( *this ); }

  RCC_Angle bodyDir() const { return  body_orientation; }
  RCC_Angle bodyDir( const RCC_Angle &dir ) { return ( body_orientation = dir ); }

  RCC_Angle headDir() const { return head_orientation; }
  RCC_Angle headDir( const RCC_Angle &dir ) { return ( head_orientation = dir ); }

  RCC_PlayerName playerName () const { return p_name; }

  virtual std::ostream& print ( std::ostream& o, RCC_Vector2d::mode_t = RCC_Vector2d::POLAR, RCC_Angle::mode_t = RCC_Angle::DEG ) const;

private:
  RCC_Angle body_orientation;
  RCC_Angle head_orientation;

  RCC_PlayerName p_name;
};

#endif
